program comp_integral_4d
!! #############################################################################
!!
!! Copyright 2021 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _comp_integral_4d:
!
! ================
! comp_integral_4d
! ================
!
! Authors
! =======
!
! Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 03/05/2021
!
! Purpose
! =======
!
! Perform vertical integrations (e.g. through the depth of the ocean or the height of the
! atmosphere) from a fourdimensional variable extracted from a NetCDF dataset.
!
! Different options are available for computing the vertical integrals in the selected
! domain and for each time steps of the fourdimensional variable (see the
! description of the **-d=** argument below).
!
! The computed integrals are stored in an output NetCDF dataset.
!
! If the NetCDF variable has missing values, in addition to those associated with a constant
! atmosphere or ocean mask, use :ref:`comp_integral_miss_4d` instead of comp_integral_4d.
!
! This procedure is parallelized if OpenMP is used and the NCSTAT software has been built
! with the _PARALLEL_READ CPP flag. Moreover, this procedure computes the vertical integrals with only 
! one pass through the data and an out-of-core strategy which is efficient on huge datasets.
!
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!
!       $ comp_integral_4d \
!         -f=input_netcdf_file \
!         -v=input_netcdf_variable \
!         -m=input_mesh_mask_netcdf_file \
!         -o=output_netcdf_file \
!         -n=output_netcdf_variable \
!         -g=grid_type                         (optional : n, t, u, v, w, f) \
!         -x=lon1,lon2                         (optional) \
!         -y=lat1,lat2                         (optional) \
!         -z=level1,level2                     (optional) \
!         -t=time1,time2                       (optional) \
!         -d=type_of_integral                  (optional : dist3, ident) \
!         -mi=missing_value                    (optional) \
!         -double                              (optional) \
!         -compact                             (optional) \
!         -bigfile                             (optional) \
!         -hdf5                                (optional) \
!         -tlimited                            (optional)
!
! By default
! ----------
!
!        *-g=*
!         the *grid_type* is set to ``n``, which means that the 3-D grid-mesh associated with
!         the input NetCDF variable is assumed to be regular or Gaussian
!        *-x=*
!         the whole longitude domain associated with the *netcdf_variable*
!        *-y=*
!         the whole latitude domain associated with the *netcdf_variable*
!        *-z=*
!         the whole vertical resolution associated with the *netcdf_variable*
!        *-t=*
!         the whole time period associated with the *netcdf_variable*
!        *-d=*
!         the *type_of_integral* is set to ``dist3``. This means that the vertical integrals are computed as
!         weighted averages and that the weight associated with each layer is proportional to the volume or the
!         mass associated with it
!        *-mi=*
!         by default, the *missing_value* in the output NetCDF variable is set to ``1.e+20``
!        *-double*
!         the integrals are stored as single-precision floating
!         point numbers in the output NetCDF file. If **-double** is activated, the
!         integrals are stored as double-precision floating point numbers 
!        *-compact*
!         the output NetCDF file is not compacted
!        *-bigfile*
!         a NetCDF classical format file is created. If **-bigfile** is activated,
!         the output NetCDF file is a 64-bit offset format file
!        *-hdf5*
!         a NetCDF classical format file is created. If **-hdf5** is activated,
!         the output NetCDF file is a NetCDF-4/HDF5 format file
!        *-tlimited*
!         the time dimension is defined as unlimited in the output NetCDF file.
!         However, if **-tlimited** is activated, the time dimension is defined
!         as limited in the output NetCDF file
!
! Remarks
! -------
!
!   #) The **-v=**\ *netcdf_variable* argument specifies the NetCDF variable from which the vertical integrals
!      must be computed and the **-f=**\ *input_netcdf_file* argument specifies that this NetCDF variable must
!      be extracted from the NetCDF file, *input_netcdf_file*.
!
!   #) If the **-x=**\ *lon1,lon2*, **-y=**\ *lat1,lat2* and **-z=**\ *level1,level2* arguments
!      are missing the whole geographical domain associated with the *netcdf_variable* is
!      used for computing the vertical integrals.
!
!      The longitude, latitude or level range must be a vector of two integers specifying the first
!      and last selected indices along each dimension. The indices are relative to ``1``\ .
!      Negative values are not allowed for *lon1*.
!
!      The **-z=** argument is used for determining the limits of the vertical integrations, which are performed.
!
!      Refer to :ref:`comp_mask_4d` for transforming geographical coordinates as indices before
!      using comp_integral_4d.
!
!   #) If the **-t=**\ *time1,time2* argument is missing the whole time period
!      associated with the *netcdf_variable* is used for computing the vertical integrals.
!
!      The selected time period is a vector of two integers specifying the first
!      and last time observations. The indices are relative to ``1``\ . Note that the
!      output NetCDF file will have ``ntime`` = *time2* - *time1* + ``1`` time observations.
!
!   #) It is assumed that the data has no missing values excepted those associated with a constant
!      atmosphere or ocean mask (as specified by the input mesh-mask NetCDF file).
!
!   #) If **-g=** is set to ``t``, ``u``, ``v``, ``w`` or ``f``, it is assumed that the NetCDF variable
!      is from an experiment with the NEMO or ORCA model.
!
!      If **-g=** is set to ``n``, it is assumed that the 3-D grid-mesh is regular or Gaussian.
!
!   #) The geographical shapes of the *netcdf_variable* (in the *input_netcdf_file*),
!      the mask (*input_mesh_mask_netcdf_file*), the scale factors (in the *input_mesh_mask_netcdf_file*)
!      must agree.
!
!   #) The **-d=** argument specifies the weighting method for computing the integrals. If:
!
!      * **-d=**\ ``dist3``, the vertical integrals are computed as weighted averages. Each layer
!        is weighted accordingly to the volume or weight associated with it. In this case, it is assumed
!        that the input mesh-mask NetCDF file has a variable named "e3"//grid, where grid is the value of
!        the **-g=** argument, which is the third scale factor associated with the grid-mesh of the
!        input_netcdf_variable. In the case of a regular or gaussian grid, this third scale factor is
!        automatically computed by :ref:`comp_clim_4d`, :ref:`comp_clim_miss_4d` or :ref:`comp_stat_4d`.
!        This third scale factor variable provides the weights for computing the vertical integrals in
!        comp_integral_miss_4d if **-d=**\ ``dist3`` is specified.
!
!      * **-d=**\ ``ident``, the vertical integrals are computed as simple sums. Each layer has the same
!        weight (equals to 1) when computing the integrals. In this case, it is not required that the
!        input mesh-mask NetCDF file contains a third scale factor as a variable.
!
!   #) The **-n=**\ *output_netcdf_variable* argument specifies the NetCDF variable which will contain
!      the computed vertical integrals in the output NetCDF file, *output_netcdf_file*, specified by the
!      **-o=**\  argument.
!
!   #) The **-mi=**\ *missing_value* argument specifies the missing value indicator associated with
!      the *output_netcdf_variable* in the *output_netcdf_file*.
!      If the **-mi=** argument is not specified, *missing_value* is set to ``1.e+20`` in the output NetCDF
!      dataset.
!
!   #) The **-double** argument specify that the interals are stored as double-precision floating
!      point numbers in the output NetCDF file.
!      By default, the vertical integrals are stored as single-precision floating point numbers
!      in the output NetCDF file.
!
!   #) If the **-compact** argument is specified and a domain is selected (with the **-x=** and **-y=**
!      arguments) then only vertical integrals for the selected domain will be output. By default, the whole
!      2-D grid is stored (with missing values outside the selected domain) in the output NetCDF file.
!
!   #) The **-bigfile** argument is allowed only if the NCSTAT software has been compiled with the
!      _USE_NETCDF36 or _USE_NETCDF4 macros (e.g. ``-D_USE_NETCDF36`` or ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 3.6 library or higher.
!      If this argument is specified, the *output_netcdf_file* will be a 64-bit offset format file
!      instead of a NetCDF classic format file. However, this argument is recognized in the procedure
!      only if the NCSTAT software has been built with the _USE_NETCDF36 or _USE_NETCDF4 CPP macros.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 macro (e.g. ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!      If this argument is specified, the *output_netcdf_file* will be a NetCDF-4/HDF5
!      format file instead of a NetCDF classic format file. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence of a parameter
!      which will be used for the computations.
!
! Outputs
! -------
!
!   comp_integral_miss_4d creates an output NetCDF file that contains the vertical integrals computed for each time step
!   and grid point selected from the input NetCDF variable and the horizontal coordinate NetCDF variables of
!   the input NetCDF dataset *input_netcdf_file* .
!   This NetCDF variable storing the integrals will have the same horizontal dimensions as the input NetCDF variable
!   in the file *input_netcdf_file* (in the description below, nlat and nlon are the lengths
!   of the spatial dimensions of the input NetCDF variable) and *ntime* time observations
!   (corresponding to the number of time steps selected with **-t=** argument) :
!
!   #) *output_netcdf_variable* ``(ntime,nlat,nlon)`` : the computed vertical integrals for the select domain
!      and time steps associated with the input NetCDF variable.
!
!   By default, the whole grid associated with the input NetCDF variable is stored (with missing values outside
!   the selected domain). Note, however, that if the argument **-compact** is used, the geographical dimensions of
!   the output NetCDF variable will be reduced to the selected domain as specified by the **-x=** and **-y=**
!   arguments (e.g., in this case ``nlat``\ =\ *lat2*\ -\ *lat1*\ +\ ``1``\ and 
!   ``nlon``\ =\ *lon2*\ -\ *lon1*\ +\ ``1``  ).
!   The number of time steps written in the output NetCDF file (e.g. ``ntime``\ ) is determined from the **-t=** argument.
!
! Examples
! --------
!
!   #)   For computing vertical integrals from the file ``ST7_1m_0101_20012_grid_T_votemper.nc``,
!        which includes a fourdimensional NetCDF variable ``votemper``,
!        and store the results in a NetCDF variable named ``int_votemper`` in the file
!        ``integral_temp.orca2.nc``, use the following command (note that the vertical integrals
!        are computed as weighted averages for each time step and grid-point associated
!        with the whole horizontal 2-D grid-mesh of the input NetCDF variable) :
!
!        ::
!
!          $ comp_integral_4d \
!            -f=ST7_1m_0101_20012_grid_T_votemper.nc \
!            -v=votemper \
!            -m=meshmask.orca2.nc \
!            -o=integral_temp.orca2.nc \
!            -n=int_votemper \
!            -g=t \
!            -d=dist3
!
!-
! ================================================================================================
!
#ifndef _OPENMP
#undef _PARALLEL_READ
#endif
!
#ifdef _USE_NETCDF4
#define _USE_NETCDF36
#endif
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env,     only : iargc, getarg
#endif
!
#ifdef _PARALLEL_READ
    use netcdf,    only : nf90_share
    use Statpack,  only : omp_limit
    use omp_lib,   only : omp_get_num_procs, omp_get_max_threads
#endif
!
#ifdef _USE_NETCDF36
    use netcdf,    only : nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,    only : nf90_hdf5
#endif
!
    use netcdf,           only : nf90_max_var_dims, nf90_max_name, nf90_real4, nf90_real8, nf90_int,          &
                                 nf90_sizehint_default, nf90_nowrite, nf90_open, nf90_noerr, nf90_strerror,   &
                                 nf90_inq_varid, nf90_inquire_variable, nf90_inquire_dimension, nf90_get_att, &
                                 nf90_get_var, nf90_create, nf90_clobber, nf90_set_fill, nf90_nofill,         &
                                 nf90_unlimited, nf90_def_dim, nf90_def_var, nf90_inq_attname, nf90_copy_att, &
                                 nf90_put_att, nf90_global, nf90_enddef, nf90_put_var, nf90_close
!
    use Statpack,         only : i1b, i4b, stnd, extd, lgl, true, false, zero, one, kfix, kexp,      &
                                 kint, tolower, case_change, string_to_val,                          &
                                 allocate_error, merror, arth
!
    use NCSTAT_Utilities, only : prtunit, nchar, sp, dp, def_missing, get_listarg, get_variable_dims
!
!
! STRONG TYPING IMPOSED 
! =====================
!
    implicit none
!
!
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='comp_integral_4d : ',   &
!
                                   cmethod(2)=(/"dist3","ident"/) 
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid, ncid_mask, ncid_out,        &
                     varid, maskid, e3id, integralid,                 &
                     i, j, j1, j2, k, nlon, nlat, nlevel, nt,        &
                     lon1, lon2, lon3, lat1, lat2, lat3,             &
                     time1, time2, time3, level1, level2, level3,    &
                     method, rec, offset_rec, natts, output_type,  &
                     output_mode
    integer(i4b), dimension(3)                 :: coor_natts, coor_type, coor_id, coor_id2
    integer(i4b), dimension(4)                 :: startlec, countlec
    integer(i4b), dimension(nf90_max_var_dims) :: dimids, dimlens
    integer(i1b)                               :: kcode
!
    real(stnd), dimension(:,:,:,:), allocatable  :: mean, std
    real(stnd), dimension(:,:,:),   allocatable  :: temp, integral, weight
    real(stnd), dimension(:),       allocatable  :: temp2
    real(stnd)                                   :: tmp, missing, scalfac, offset
    real(extd), dimension(:),       allocatable  :: record
!
    logical(lgl)                                 :: present_missing, use_tlimited, create_hdf5,    &
                                                    use_x, use_y, use_z, use_compact, add_offset, scale_factor
    logical(lgl), dimension(3)                   :: use_coor_variable
    logical(lgl), dimension(:,:,:), allocatable  :: mask
    logical(lgl), dimension(:,:),   allocatable  :: mask_integral
!
    character(len=nchar)                       :: name, history
    character(len=nf90_max_name)               :: fic_name, var_name, mask_name,    &
                                                  out_name, int_name, att_name 
    character(len=nf90_max_name), dimension(4) :: dim_names
    character(len=14)                          :: fmt
    character                                  :: chr, grid_type
!
#ifdef _PARALLEL_READ
    integer(i4b) :: ncid2, varid2
!
    logical  :: test_par
!
#endif
!
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    if ( narg==0 ) then
!
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Perform vertical integrations from a fourdimensional variable '
        write (prtunit,*) ' extracted from a NetCDF dataset and write the results in '
        write (prtunit,*) ' a NetCDF dataset'
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' comp_integral_4d -f=input_netcdf_file'
        write (prtunit,*) '                   -v=input_netcdf_variable'
        write (prtunit,*) '                   -m=input_mesh_mask_netcdf_file'
        write (prtunit,*) '                   -o=output_netcdf_file'
        write (prtunit,*) '                   -n=output_netcdf_variable'
        write (prtunit,*) '                   -g=grid_type                         (optional : n, t, u, v, w, f)'
        write (prtunit,*) '                   -x=lon1,lon2                         (optional)'
        write (prtunit,*) '                   -y=lat1,lat2                         (optional)'
        write (prtunit,*) '                   -z=level1,level2                     (optional)'
        write (prtunit,*) '                   -t=time1,time2                       (optional)'
        write (prtunit,*) '                   -d=type_of_integral                  (optional : dist3, ident)'
        write (prtunit,*) '                   -mi=missing_value                    (optional)'
        write (prtunit,*) '                   -double                              (optional)'
        write (prtunit,*) '                   -compact                             (optional)'
#ifdef _USE_NETCDF36
        write (prtunit,*) '                   -bigfile                             (optional)'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '                   -hdf5                                (optional)'
#endif
        write (prtunit,*) '                   -tlimited                            (optional)'
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-g=n'
        write (prtunit,*) '-x= the whole longitude domain associated with the input_netcdf_variable'
        write (prtunit,*) '-y= the whole latitude domain associated with the input_netcdf_variable'
        write (prtunit,*) '-z= the whole vertical resolution associated with the input_netcdf_variable'
        write (prtunit,*) '-t= the whole time period  associated with the input_netcdf_variable'
        write (prtunit,*) '-d=dist3'
        write (prtunit,*) '-mi= the missing_value in the output NetCDF file is set to 1.e+20'
        write (prtunit,*) '-double   : the results are stored as single floating numbers'
        write (prtunit,*) '-compact  : by default, the output NetCDF file is not compacted'
#ifdef _USE_NETCDF36
        write (prtunit,*) '-bigfile  : a netCDF classical format file is created'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5     : a NetCDF classical format file is created'
#endif
        write (prtunit,*) '-tlimited : the time dimension is defined as unlimited'
        write (prtunit,*)
#ifdef _PARALLEL_READ
        write (prtunit,*) 'This procedure is parallelized with OpenMP'
#endif
        write (prtunit,*)
!
        stop
!
    else if ( narg<5 .or. narg>17 ) then
!
        call merror( name_proc//"Wrong number of line arguments" )
!
    end if
!
    fic_name  = " "
    var_name  = " "
    mask_name = " "
    out_name  = " "
    int_name  = " "
!
    history = "comp_integral_4d"
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    output_mode   = nf90_clobber
    output_type   = nf90_real4
    method      = 1
!
    missing       = def_missing
!
    grid_type     = "n"
!
    lon1      = 1
    lon2      = huge( lon2 )
    lat1      = 1
    lat2      = huge( lat2 )
    level1    = 1
    level2    = huge( level2 )
    time1     = 1
    time2     = huge( time2 )
!
    use_x        = false
    use_y        = false
    use_z        = false
    use_compact  = false
    create_hdf5  = false
    use_tlimited = false
!
!   GET THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
!
        history = trim(history)//" "//trim(name)
!
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-f=')
!
!               GET THE NAME OF THE NETCDF DATASET.
!
                fic_name = name(4:)
!
            case ('-v=')
!
!               GET THE NAME OF THE NETCDF VARIABLE IN THE DATASET.
!
                var_name = name(4:)
!
            case ('-g=')
!
!               GET THE GRID-TYPE.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( 'n', 'u', 'v', 'w', 't', 'f' )
                        grid_type = name(4:4)
                    case default
                        call merror( name_proc//"Unknown grid type" )
                end select
!
            case ('-x=')
!
!               GET THE DOMAIN DEFINITION (Longitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lon1 and lon2 are integer values",     &
                                  "Wrong domain definition (longitude)",  &
                                  lon1, lon2 )
!
                use_x = true
!
            case ('-y=')
!
!               GET THE DOMAIN DEFINITION (latitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lat1 and lat2 are integer values",     &
                                  "Wrong domain definition (latitude)",   &
                                  lat1, lat2 )
!
                use_y = true
!
            case ('-z=')
!
!               GET THE LEVEL DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "level1 and level2 are integer values", &
                                  "Wrong level definition",               &
                                  level1, level2 )
!
                use_z = true
!
            case ('-t=')
!
!               GET THE PERIOD DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "time1 and time2 are integer values",   &
                                  "Wrong period definition",              &
                                  time1, time2 )
!
            case ('-m=')
!
!               GET THE NAME OF THE NETCDF MESH-MASK DATASET.
!
                mask_name = name(4:)
!
            case ('-d=')
!
!               DETERMINE THE METHOD FOR THE VERTICAL INTEGRATION.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( cmethod(1) )
                        method = 1
                    case ( cmethod(2) )
                        method = 2
                    case default
                        call merror( name_proc//"Unknown integration method" )
                end select
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR STORING THE VERTICAL INTEGRALS.
!
               out_name = name(4:)
!
            case ('-n=')
!
!               GET THE NAME OF THE NEW VARIABLE STORING THE VERTICAL INTEGRALS.
!
                int_name = name(4:)
!
            case default
!
                select case ( name(1:4 ) )
!
                    case ('-mi=')
!
!                       GET THE MISSING VALUE INDICATOR.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode/=kfix .and. kcode/=kexp )                &
                    call merror( name_proc//"The -mi= argument must be a real number " )
                        read( name(5:), fmt ) missing
!
                    case default
!
                        call case_change( name(4:), tolower )
!
                        select case ( name(1:) )
!
                            case ('-tlimited')
!
!                               DEFINE THE TIME DIMENSION AS LIMITED.
!
                                use_tlimited = true
!
                            case ('-double')
!
!                               STORE RESULTS AS DOUBLE PRECISION NUMBERS.
!
                                output_type = nf90_real8
!
#ifdef _USE_NETCDF4
                            case ('-hdf5')
!
!                               REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                output_mode = nf90_hdf5
                                create_hdf5 = true
#endif
!
#ifdef _USE_NETCDF36
                            case ('-bigfile')
!
!                               REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                output_mode = nf90_64bit_offset
#endif
!
                            case ('-compact')
!
!                               COMPACT THE OUTPUT FILE.
!
                                use_compact = true
!
                            case default
!
!                               WRONG ARGUMENT.
!
                                call merror( name_proc//"Wrong argument" )
!
                        end select
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    if ( fic_name==" " ) call merror( name_proc//"Input NETCDF file is missing" )
!
    if ( var_name==" " ) call merror( name_proc//"Input NETCDF variable is missing" )
!
    if ( mask_name==" " ) call merror( name_proc//"Input mesh-mask NETCDF file is missing" )
!
    if ( out_name==" " ) call merror( name_proc//"Output NETCDF file is missing" )
!
    if ( int_name==" " ) call merror( name_proc//"Output NETCDF variable is missing" )
!
!   OPEN THE NETCDF DATASETS.
!
    k = nf90_sizehint_default
    status = nf90_open( trim(fic_name), nf90_nowrite, ncid, chunksize=k )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_open( trim(mask_name), nf90_nowrite, ncid_mask )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   GET THE VARIABLE ID FROM ITS NAME.
!
    status = nf90_inq_varid( ncid, trim(var_name), varid )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   FIND THE EXTERNAL TYPE, THE DIMENSION LENGTHS AND NAMES AND THE NUMBER OF ATTRIBUTES
!   FOR THE NETCDF VARIABLE.
!
    status = nf90_inquire_variable( ncid, varid, ndims=i, dimids=dimids, natts=natts )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( i/=4 ) call merror( name_proc//"Input NETCDF variable is not fourdimensional" )
!
    do i = 1, 4
!
        status = nf90_inquire_dimension( ncid, dimids(i), name=dim_names(i), len=dimlens(i) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
    nlon   = dimlens(1)
    nlat   = dimlens(2)
    nlevel = dimlens(3)
    nt     = dimlens(4)
!
!   FIND THE SCALE_FACTOR ATTRIBUTE ASSOCIATED WITH THE NETCDF VARIABLE, IF ANY.
!
    status = nf90_get_att( ncid, varid, "scale_factor", scalfac )
    if ( status/=nf90_noerr ) then
        scale_factor = false
        scalfac      = one
    else
        scale_factor = true
    end if
!
!   FIND THE ADD_OFFSET ATTRIBUTE ASSOCIATED WITH THE NETCDF VARIABLE, IF ANY.
!
    status = nf90_get_att( ncid, varid, "add_offset", offset )
    if ( status/=nf90_noerr ) then
        add_offset = false
        offset     = zero
    else
        add_offset = true
    end if
!
    j = 0
!
    do i = 1, 4
!
        if ( i==3 ) cycle
!
        j = j + 1
!
!       GET THE COORDINATE VARIABLES IDS FROM THEIR NAME, IF ANY.
!
        status = nf90_inq_varid( ncid, trim(dim_names(i)), coor_id(j) )
        if ( status==nf90_noerr ) then
!
!           FIND THE EXTERNAL TYPE, THE DIMENSION ID AND THE NUMBER OF ATTRIBUTES 
!           FOR THE COORDINATE VARIABLES.
!
            status = nf90_inquire_variable( ncid, coor_id(j), xtype=coor_type(j), ndims=k,    &
                                            dimids=dimlens, natts=coor_natts(j) )
            if ( status/=nf90_noerr )                                                         &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            use_coor_variable(j) = k==1 .and. dimlens(1)==dimids(i)
        else
            use_coor_variable(j) = false
        end if
!
        if ( .not.use_coor_variable(j) ) coor_type(j) = nf90_int
!
    end do
!
    chr = merge( 't', grid_type, grid_type=='w' )
!
!   GET THE MASK ID FROM ITS NAME.
!   VERIFY THE DIMENSION LENGTHS FOR THE NETCDF MASK VARIABLE.
!
    call get_variable_dims( name_proc, chr//"mask", ncid_mask, maskid, dimids, dimlens )
!
    if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat .or. dimlens(3)/=nlevel )    &
    call merror( name_proc//"Shape disagreement with the mask" )
!
!   GET THE LONGITUDE BOUNDS, IF REQUIRED.
!
    if ( .not.use_x ) then
!
        status = nf90_get_att( ncid_mask, maskid, "lon1_Eastern_limit", i )
        if ( status==nf90_noerr ) then
!
            if ( i>0 ) then
!
                lon1 = i
!
                status = nf90_get_att( ncid_mask, maskid, "lon2_Western_limit", i )
                if ( status==nf90_noerr ) lon2 = i
!
            end if
!
        end if
!
    end if
!
!   GET THE LATITUDE BOUNDS, IF REQUIRED.
!
    if ( .not.use_y ) then
!
        status = nf90_get_att( ncid_mask, maskid, "lat1_Southern_limit", i )
        if ( status==nf90_noerr ) lat1 = i
!
        status = nf90_get_att( ncid_mask, maskid, "lat2_Northern_limit", i )
        if ( status==nf90_noerr ) lat2 = i
!
    end if
!
!   GET THE LEVEL BOUNDS, IF REQUIRED.
!
    if ( .not.use_z ) then
!
        status = nf90_get_att( ncid_mask, maskid, "level1_First_level", i )
        if ( status==nf90_noerr ) level1 = i
!
        status = nf90_get_att( ncid_mask, maskid, "level2_Last_level", i )
        if ( status==nf90_noerr ) level2 = i
!
    end if
!
!   CHECK THE BOUNDARIES OF THE GEOGRAPHICAL DOMAIN.
!
    lat1 = max( lat1, 1 )
    lat2 = min( lat2, nlat )
    lat3 = lat2 - lat1 + 1
    if ( lat3<=0  )  call merror( name_proc//"Wrong domain definition (latitude)" )
!
    if ( lon1<=0  )  call merror( name_proc//"Wrong domain definition (longitude)" )
    lon2 = min( lon2, nlon )
    lon3 = lon2 - lon1 + 1
    if ( lon3<=0  )  call merror( name_proc//"Wrong domain definition (longitude)" )
!
!   CHECK THE VERTICAL DEFINITION.
!
    level1 = max( level1, 1 )
    level2 = min( level2, nlevel )
    level3 = level2 - level1 + 1
    if ( level3<=0  ) call merror( name_proc//"Wrong level definition" )
!
!   CHECK THE PERIOD DEFINITION.
!
    time1 = max( time1, 1 )
    time2 = min( time2, nt )
    time3 = time2 - time1 + 1
    if ( time3<=0  ) call merror( name_proc//"Wrong time definition" )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( temp(lon3,lat3,level3), weight(lon3,lat3,level3), mask(lon3,lat3,level3),  &
              mask_integral(lon3,lat3), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
    startlec(1:4) = (/ lon1, lat1, level1, time1 /)
    countlec(1:4) = (/ lon3, lat3, level3, 1_i4b /)
!
!   GET THE MASK.
!
    status = nf90_get_var( ncid_mask, maskid, temp, start=startlec(1:3), count=countlec(1:3) )     
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   DETERMINE THE GRID POINTS AND LEVELS USED FOR VERTICAL INTEGRATION FROM THE MASK.
!
    mask(:lon3,:lat3,:level3) = temp(:lon3,:lat3,:level3)==one
!
    mask_integral(:lon3,:lat3) = any( mask(:lon3,:lat3,:level3), dim=3 )
    i = count( mask_integral(:lon3,:lat3) )
!
    if ( i<=0  )  &
    call merror( name_proc//"Wrong domain or levels definitions : integrals can not be computed" )
!
    present_missing = i/=lon3*lat3
!
    if ( present_missing ) then
!
        mask_integral(:lon3,:lat3) = .not.mask_integral(:lon3,:lat3)
!
    end if
!
!   SELECT THE METHOD OF VERTICAL INTEGRATION.
!
    if ( method==1 ) then
!
!       THE VERTICAL INTEGRAL IS COMPUTED AS A WEIGHTED AVERAGE.
!
!       ALLOCATE WORK ARRAY.
!
        allocate( temp2(level3), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
        chr = merge( grid_type, 't', grid_type=='w' .or. grid_type=='n' )
!
!       GET THE THIRD SCALE FACTOR ID FROM ITS NAME.
!       VERIFY THE DIMENSION LENGTHS FOR THE THIRD SCALE FACTOR VARIABLE.
!
        call get_variable_dims( name_proc, "e3"//chr, ncid_mask,                    &
                                e3id, dimids, dimlens )
!
        if ( dimlens(1)/=1 .or. dimlens(2)/=1 .or. dimlens(3)/=nlevel )             &
        call merror( name_proc//"Shape disagreement with the third scale factor" )
!
!       GET THE THIRD SCALE FACTOR VARIABLE.
!
        status = nf90_get_var( ncid_mask, e3id, temp2(:level3),                     &
                         start=(/1_i4b,1_i4b,level1/), count=(/1_i4b,1_i4b,level3/) )    
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       CHECK THE THIRD SCALE FACTOR.
!
        if ( temp2(1)<zero )                        &
    call merror( name_proc//" First level with -d=dist3 are not allowed for mesh-mask "//trim(mask_name) )
!
        if ( temp2(level3)<zero )                   &
    call merror( name_proc//" Last level with -d=dist3 are not allowed for mesh-mask "//trim(mask_name) )
!
!       MULTIPLY THE WEIGHTS AND THE MASK.
!
        do i = 1, level3
!
            weight(:lon3,:lat3,i) = temp(:lon3,:lat3,i)*temp2(i)
!
        end do
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( temp2 )
!
    end if
!
!   ALLOCATE WORK ARRAY FOR STORING THE VERTICAL INTEGRALS.
!
    allocate( integral(lon3,lat3,time3), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
#ifdef _PARALLEL_READ
!
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    test_par = i>1 .and. j>1 .and. time3*lon3*lat3*level3>omp_limit
!
    if ( test_par ) then
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( temp )
    end if
!
#endif
!
!   COMPUTE THE OFFSET FOR THE RECORD NUMBER.
!
    offset_rec = time1 - 1
!
!   COMPUTE THE VERTICAL INTEGRALS FOR EACH TIME STEP.
!
#ifdef _PARALLEL_READ
!
    if ( test_par ) then
!
!$OMP PARALLEL PRIVATE(k,status,startlec,ncid2,varid2,temp)                  &
!$OMP         ,SHARED(lon1,lon3,lat1,lat3,level1,level3,time3,offset_rec,    &
!$OMP                 varid,countlec,weight,integral,scalfac,offset)
!
!       ALLOCATE WORK ARRAY.
!
        allocate( temp(lon3,lat3,level3), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
!$OMP CRITICAL (opennc)
!
!       OPEN THE NETCDF FILE.
!
        status = nf90_open( trim(fic_name), nf90_share, ncid2  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       GET THE VARIABLE ID FROM ITS NAME.
!
        status = nf90_inq_varid( ncid2, trim(var_name), varid2 )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (opennc)
!
!       READ THE DATA.
!
        startlec(1)  = lon1
        startlec(2)  = lat1
        startlec(3)  = level1
!
!$OMP DO SCHEDULE(STATIC)
        do k = 1, time3
!
            startlec(4) = k + offset_rec
!
!           FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE.
!
            status = nf90_get_var( ncid2, varid2, temp(:lon3,:lat3,:level3),                   &
                                   start=startlec, count=countlec )     
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           MULTIPLY THE NETCDF VARIABLE BY THE SCALE FACTOR IF REQUIRED.
!
            if ( scale_factor ) then
!
                where ( mask(:lon3,:lat3,:level3) )     &
                     temp(:lon3,:lat3,:level3) = temp(:lon3,:lat3,:level3)*scalfac
!
            end if
!
!           ADD OFFSET TO THE NETCDF VARIABLE IF REQUIRED.
!
            if ( add_offset ) then
!
                where ( mask(:lon3,:lat3,:level3) )     &
                     temp(:lon3,:lat3,:level3) = temp(:lon3,:lat3,:level3) + offset
!
            end if
!
!           ADJUST FOR THE WEIGHTS AND MASK.
!
            temp(:lon3,:lat3,:level3) = temp(:lon3,:lat3,:level3)*weight(:lon3,:lat3,:level3)
!
!           COMPUTE THE VERTICAL INTEGRALS.
!
            integral(:lon3,:lat3,k) = sum( temp(:lon3,:lat3,:level3), dim=3 )
!
            if ( present_missing ) then
!
                where( mask_integral(:lon3,:lat3) ) integral(:lon3,:lat3,k) = missing
!
            end if
!
        end do
!$OMP END DO
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( temp )
!
!$OMP CRITICAL (closenc)
!
!       CLOSE THE NETCDF DATASET.
!
        status = nf90_close( ncid2 )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (closenc)
!
!$OMP END PARALLEL
!
    else
!
#endif
!
        do k = 1, time3
!
            startlec(4) = k + offset_rec
!
!           FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE.
!
            status = nf90_get_var( ncid, varid, temp(:lon3,:lat3,:level3),                    &
                                   start=startlec, count=countlec )     
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           MULTIPLY THE NETCDF VARIABLE BY THE SCALE FACTOR IF REQUIRED.
!
            if ( scale_factor ) then
!
                where ( mask(:lon3,:lat3,:level3) )     &
                     temp(:lon3,:lat3,:level3) = temp(:lon3,:lat3,:level3)*scalfac
!
            end if
!
!           ADD OFFSET TO THE NETCDF VARIABLE IF REQUIRED.
!
            if ( add_offset ) then
!
                where ( mask(:lon3,:lat3,:level3) )     &
                     temp(:lon3,:lat3,:level3) = temp(:lon3,:lat3,:level3) + offset
!
            end if
!
!           ADJUST FOR THE WEIGHTS AND MASK.
!
            temp(:lon3,:lat3,:level3) = temp(:lon3,:lat3,:level3)*weight(:lon3,:lat3,:level3)
!
!           COMPUTE THE VERTICAL INTEGRALS.
!
            integral(:lon3,:lat3,k) = sum( temp(:lon3,:lat3,:level3), dim=3 )
!
            if ( present_missing ) then
!
                where( mask_integral(:lon3,:lat3) ) integral(:lon3,:lat3,k) = missing
!
            end if
!
        end do
!
#ifdef _PARALLEL_READ
!
    end if
!
#endif
!
!   CREATE THE NEW NETCDF DATASET FOR THE INTEGRAL DATA.
!
    if ( create_hdf5 ) then
        status = nf90_create( trim(out_name), cmode=output_mode, ncid=ncid_out )
    else
        k = nf90_sizehint_default
        status = nf90_create( trim(out_name), cmode=output_mode, ncid=ncid_out, chunksize=k )
    end if
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( use_compact ) then
        status = nf90_set_fill( ncid_out, nf90_nofill, i )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end if
!
!   DEFINE THE DIMENSIONS FOR THE NEW NETCDF DATASET.
!
    if ( use_compact ) then
        startlec(1:3) = (/ lon1, lat1, time1 /)
        dimlens(1:2)  = (/ lon3, lat3 /)
    else
        startlec(1:3) = (/ 1_i4b, 1_i4b, time1 /)
        dimlens(1:2)  = (/ nlon, nlat /)
    end if
!
    if ( use_tlimited ) then
        dimlens(3)   = time3
    else
        dimlens(3)   = nf90_unlimited
    end if
!
    dim_names(3) = dim_names(4)
!
    do i = 1, 3
!
        status = nf90_def_dim( ncid_out, trim(dim_names(i)), len=dimlens(i), dimid=dimids(i) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
!   DEFINE THE VARIABLES FOR THE NEW NETCDF DATASET.
!
    do i = 1, 3
!
        status = nf90_def_var( ncid_out, trim(dim_names(i)), coor_type(i),                &
                               dimids=dimids(i), varid=coor_id2(i) )
        if ( status/=nf90_noerr )                                                         &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
    status = nf90_def_var( ncid_out, int_name, output_type, dimids=dimids(:3), varid=integralid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   ASSIGN ATTRIBUTES FOR THE NEW NETCDF DATASET.
!
!   COPY ATTRIBUTES FOR COORDINATE VARIABLES.
!
    do j = 1, 3
!
        if ( .not.use_coor_variable(j) ) cycle
!
        do i = 1, coor_natts(j)
!
            status = nf90_inq_attname( ncid, coor_id(j), i, att_name )
            if ( status/=nf90_noerr )                                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           COPY ATTRIBUTES FOR THE COORDINATE VARIABLES.
!
            status = nf90_copy_att(  ncid, coor_id(j), att_name, ncid_out, coor_id2(j) )
            if ( status/=nf90_noerr )                                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    end do
!
    if ( .not.use_coor_variable(3) ) then
!
        status = nf90_put_att( ncid_out, coor_id2(3), "long_name", "Time axis" )
        if ( status/=nf90_noerr )                                                  &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_out, coor_id2(3), "short_name", "Time" )
        if ( status/=nf90_noerr )                                                  &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
!   COPY ATTRIBUTES FOR THE INTEGRAL VARIABLE.
!
    do i = 1, natts
!
        status = nf90_inq_attname( ncid, varid, i, att_name )
        if ( status/=nf90_noerr )  call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        if ( att_name=="add_offset"         .or.  &
             att_name=="scale_factor"       .or.  &
             att_name=="missing_value"      .or.  &
             att_name=="units"              .or.  &
             att_name=="long_name"          .or.  &
             att_name=="short_name"         .or.  &
             att_name=="_FillValue"               )   cycle
!
!       COPY USEFUL ATTRIBUTES .
!
        status = nf90_copy_att(  ncid, varid, att_name, ncid_out, integralid )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
    status = nf90_put_att( ncid_out, integralid, "long_name", trim(int_name)//" vertical integral" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_out, integralid, "short_name", trim(int_name)//" vertical integral" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( method==1 ) then
!
        status = nf90_get_att( ncid, varid, 'units', name )
        if ( status==nf90_noerr ) then
            status = nf90_put_att( ncid_out, integralid, "original_units", trim(name) )
        end if
!
    else
!
        status = nf90_copy_att(  ncid, varid, 'units', ncid_out, integralid )
!
    end if
!
    if (  output_type==nf90_real8 ) then
!
        status = nf90_put_att( ncid_out, integralid, "missing_value", real( missing, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_out, integralid, "_FillValue", real( missing, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    else
!
        status = nf90_put_att( ncid_out, integralid, "missing_value", real( missing, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_out, integralid, "_FillValue", real( missing, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
!   SET GLOBAL ATTRIBUTES.
!
    status = nf90_put_att( ncid_out, nf90_global, "title", " Vertical integral for "//trim(var_name) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_out, nf90_global, "integration_method", cmethod(method) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_out, nf90_global, "grid_type", grid_type )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_out, nf90_global, "history", trim(history) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   LEAVE DEFINE MODE FOR THE NEW NETCDF DATASET.
!
    status = nf90_enddef( ncid_out )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE VARIABLES FOR THE NEW NETCDF DATASET.
!
!   WRITE COORDINATE VARIABLES FIRST.
!
    dimlens(3) = time3
!
    do j = 1, 3
!
        j2 = dimlens(j)
!
        if ( use_coor_variable(j) ) then
!
            j1 = startlec(j)
!
            allocate( record(j2), stat=status )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
            status = nf90_get_var( ncid, coor_id(j), record, start=(/ j1 /), count=(/ j2 /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_var( ncid_out, coor_id2(j), record )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            deallocate( record )
!
        else
            status = nf90_put_var( ncid_out, coor_id2(j), arth(j1,1_i4b,j2 ) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        end if
!
    end do
!
!   WRITE INTEGRAL DATA.
!
    if ( use_compact ) then
        startlec(1:2) = 1
    else
        startlec(1:2) = (/ lon1, lat1 /)
    end if
!
    countlec(1:3) = (/ lon3, lat3, 1_i4b /)
!
    do i = 1, time3
!
        startlec(3) = i
!
        status = nf90_put_var( ncid_out, integralid, integral(:lon3,:lat3,i), start=startlec(1:3), count=countlec(1:3) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( allocated(temp) ) then
        deallocate( mask, mask_integral, integral, weight, temp )
    else
        deallocate( mask, mask_integral, integral, weight )
    end if
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_close( ncid_mask )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_close( ncid_out )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!
! END OF PROGRAM comp_integral_4d
! ===============================
!
end program comp_integral_4d
